//#define DEBUG

#include <stdio.h>
#include <stdint.h>
#include <limits.h>
#include <map>
#include <list>
#include <vector>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <arpa/inet.h>

#include <unordered_map>
#include <unordered_set>
#include <string>


unsigned log2(unsigned x)
{
    unsigned i;
    for (i = 0; i < 31; i++) {
        if (x == (1U << i)) {
            return i;
        }
    }
    return 0;
}


int main(int argc, char ** argv)
{
    if (argc != 4) {
        printf("usage: directmap "
            "<trace_in file> <cache size> <line size>\n"
            "<cache size>           in bytes, e.g. 8192.\n"
            "<line size>            in bytes, e.g. 4.\n");
        return 1;
    }
    FILE * trace_in = fopen(argv[ 1 ], "rb");
    if (trace_in == NULL) {
        printf("unable to open trace_in file\n");
        return 1;
    }

    unsigned memory_size = log2(atoi(argv[ 2 ]));
    unsigned line_size = log2(atoi(argv[ 3 ]));
    if ((memory_size == 0)
    || (line_size == 0)
    || (memory_size <= line_size)) {
        printf("sizes are invalid\n");
        return 1;
    }
    memory_size -= line_size;
    unsigned tag_shift = line_size + memory_size;

    uint64_t        hit = 0;
    uint64_t        total = 0;
    uint32_t *      cache_contents = new uint32_t[1 << memory_size];
    uint32_t        address = 0;

    memset(cache_contents, 0xff, 4 << memory_size);

    while (fread(&address, 4, 1, trace_in) == 1) {
        uint32_t line = (address >> line_size) & ((1 << memory_size) - 1);
        uint32_t tag = address >> tag_shift;

        if (cache_contents[line] == tag) {
            hit++;
        } else {
            cache_contents[line] = tag;
        }
        total++;
    }
    fclose(trace_in);

    fprintf(stderr, "cache size %u\n"
                    "total %1.0f\n"
                    "misses %1.0f\n"
                    "bytes transferred %1.0f\n"
                    "miss ratio %1.5f\n",
                    1 << tag_shift,
                    (double) total, (double) (total - hit), 
                    (double) ((total - hit) * (1 << line_size)),
                    (double) (total - hit) / (double) total);
    fprintf(stdout, "%1.0f\n", 
                    (double) ((total - hit)));

    return 0;
}

