
//#define DEBUG

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <limits.h>
#include <map>
#include <list>
#include <vector>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <regex.h>

#include <unordered_map>
#include <unordered_set>
#include <set>
#include <string>

#define CALL_FLAG 1
static const uint32_t CUTOFF_ADDRESS = 0xff000000U;


struct Method
{
    Method(uint32_t _start, std::string _name, uint32_t _size) :
            start(_start), name(_name), size(_size),
            call_count(0), return_count(0) {};
    virtual ~Method() {};

    uint32_t start;
    std::string name;
    uint32_t size;
    uint32_t call_count, return_count;
    std::map<const Method *, unsigned> link_count;
};


int main(int argc, char ** argv)
{
    if (argc != 4) {
        printf("usage: opt_pure_trace "
            "<trace_in file> <nm file> <graph out file>\n");
        return 1;
    }
    const char * trace_in_file = argv[1];
    const char * nm_file = argv[2];
    const char * graph_out_file = argv[3];

    FILE * trace_in = fopen(trace_in_file, "rb");
    if (trace_in == NULL) {
        printf("unable to open trace_in file\n");
        return 1;
    }

    FILE * nm_in = fopen(nm_file, "rt");
    if (nm_in == NULL) {
        printf("unable to open nm_in file\n");
        return 1;
    }

    FILE * graph_out = fopen(graph_out_file, "wt");
    if (graph_out == NULL) {
        printf("unable to open graph_out file\n");
        return 1;
    }

    std::unordered_map<uint32_t, Method *> methods;
    std::vector<Method *> methods_list;

    // scan nm file to identify functions
    char buf[BUFSIZ];
    regex_t pr;
    regmatch_t pmatch[4];
    unsigned exit_at = 0;

    int rc = regcomp(&pr, 
            "^([0-9a-f]{8}) ([0-9a-f]{8}|) *[tT] ([^ ]*).$", REG_EXTENDED);
    assert(rc == 0);

    while (fgets(buf, BUFSIZ - 1, nm_in)) {
        if (regexec(&pr, buf, 4, pmatch, 0) == 0) {
            assert(pmatch[1].rm_so == 0);
            assert(pmatch[2].rm_so > 0);
            assert(pmatch[3].rm_so > 0);

            buf[pmatch[1].rm_eo] = '\0';
            unsigned start = strtol(&buf[pmatch[1].rm_so], NULL, 16);

            unsigned size = 0;
            if (pmatch[2].rm_so != pmatch[2].rm_eo) {
                buf[pmatch[2].rm_eo] = '\0';
                size = strtol(&buf[pmatch[2].rm_so], NULL, 16);
            }
            buf[pmatch[3].rm_eo] = '\0';
            const char * name = &buf[pmatch[3].rm_so];

            if (start && (methods.count(start) == 0)) {
                Method * m = new Method(start, name, size);
                methods[start] = m;
                methods_list.push_back(m);

                if (strcmp(name, "exit") == 0) {
                    exit_at = start;
                }
            }
        }
    }
    regfree(&pr);

    // how many addresses in the trace?
    fseek(trace_in, 0, SEEK_END);
    long end_of_file = ftell(trace_in);
    assert(end_of_file > 0);
    assert((end_of_file & 3) == 0);
    uint64_t num_addresses = end_of_file / 4;
    uint64_t i;

    std::vector<std::pair<uint32_t, uint32_t> > return_stack;

    uint32_t old_address = 0;
    uint32_t address = 0;
    uint32_t call_flag = CALL_FLAG;
    fseek(trace_in, 0, SEEK_SET);
    fread(&address, 4, 1, trace_in);
    assert(methods.count(address));
    assert(methods[address]->name == "main");
    return_stack.push_back(std::pair<uint32_t, uint32_t>(0, 0));

    fseek(trace_in, 0, SEEK_SET);

    for (i = 0; i < num_addresses; i++) {
        fread(&address, 4, 1, trace_in);
        if (address >= CUTOFF_ADDRESS) continue;

        uint32_t next_call_flag = address & CALL_FLAG;
        address &= ~3;

        if (address == exit_at) break;

        if (call_flag) {
            // call identified
            if (!methods.count(address)) {
                fprintf(stderr, "Call to unknown function at %08x\n"
                    "Source %08x\n", address, old_address);
                exit(1);
            }

            Method * child = methods[address];
            child->call_count++;

            uint32_t parent_start = return_stack.back().second;
            if (methods.count(parent_start)) {
                Method * parent = methods[parent_start];
                parent->link_count[child]++;

            }

            return_stack.push_back(
                std::pair<uint32_t, uint32_t> (
                        old_address + 4, address));
            

        } else if (return_stack.back().first == address) {
            // return identified
            uint32_t child_start = return_stack.back().second;
            Method * child = methods[child_start];

            return_stack.pop_back();
            assert(!return_stack.empty());
            assert(child);
            child->return_count++;

        } else {
#if 0
            uint32_t parent_start = return_stack.back().second;

            if (methods.count(parent_start)) {
                Method * parent = methods[parent_start];
                if ((address < parent_start)
                || (address > (parent_start + parent->size))) {
                    fprintf(stderr, "Jump out of method detected?\n"
                        "Method bounds %08x to %08x (%s)\n" 
                        "old address %08x\n"
                        "address %08x\n", 
                            parent_start, parent_start + parent->size,
                            parent->name.c_str(),
                            old_address, address);
                    fprintf(stderr, "return stack contents\n");
                    unsigned j;
                    for (j = 0; j < return_stack.size(); j ++) {
                        fprintf(stderr, 
                                "%u return address %08x   called %08x\n",
                                j, return_stack[j].first,
                                return_stack[j].second);
                    }
                    exit(1);
                }
            }
#endif
        }
        call_flag = next_call_flag;
        old_address = address;
    }

    off_t crop_at = ftell(trace_in);
    fclose(trace_in);
    fclose(nm_in);

    if (address == exit_at) {
        truncate(trace_in_file, crop_at);
        printf("cropping\n");
    }

    for (i = 0; i < methods_list.size(); i++) {
        Method * m = methods_list[i];

        if (m->call_count == 0) {
            continue;
        }
        fprintf(graph_out, "method(0x%x, '%s', 0x%x, 0x%x)\n",
                m->start, m->name.c_str(), m->size, m->call_count);


        std::map<const Method *, unsigned>::iterator li;
        for (li = m->link_count.begin();
                li != m->link_count.end(); li++) {
            fprintf(graph_out, "link(0x%x, 0x%x, 0x%x)\n",
                m->start, (* li).first->start, (* li).second);
        }
    }

#if 0
    fprintf(stderr, "return stack contents\n");
    for (i = 0; i < return_stack.size(); i ++) {
        fprintf(stderr, "return address %08x   called %08x\n",
                return_stack[i].first,
                return_stack[i].second);
    }
#endif

    fclose(graph_out);
}


